"""
Copied out of pyflakes to allow additions (denoted by "nbsmoke
addition"). Additions:

 * return more info for use by caller

 * support "# noqa" comments in ipynb (to override flakes, e.g. if unused import is deliberate)

 * support for blacklisted magics triggering flakes
"""

import sys
import re
import pyflakes.reporter, pyflakes.checker
import _ast

### nbsmoke addition ######################################################
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
###########################################################################

### nbsmoke addition ######################################################
import collections
FakeMessage = collections.namedtuple("FakeMessage", "lineno col message message_args")
###########################################################################


def flake_check(codeString, filename, reporter=None):
    if reporter is None:
        reporter = pyflakes.reporter._makeDefaultReporter()
    # First, compile into an AST and handle syntax errors.
    try:
        tree = compile(codeString, filename, "exec", _ast.PyCF_ONLY_AST)
    except SyntaxError:
        value = sys.exc_info()[1]
        msg = value.args[0]

        (lineno, offset, text) = value.lineno, value.offset, value.text

        if pyflakes.checker.PYPY:
            if text is None:
                lines = codeString.splitlines()
                if len(lines) >= lineno:
                    text = lines[lineno - 1]
                    if sys.version_info >= (3, ) and isinstance(text, bytes):
                        try:
                            text = text.decode('ascii')
                        except UnicodeDecodeError:
                            text = None
            offset -= 1

        # If there's an encoding problem with the file, the text is None.
        if text is None:
            # Avoid using msg, since for the only known case, it contains a
            # bogus message that claims the encoding the file declared was
            # unknown.
            reporter.unexpectedError(filename, 'problem decoding source')
        else:
            reporter.syntaxError(filename, msg, lineno, offset, text)
### nbsmoke addition ######################################################
# be able to report the syntax error details if "onlywarn"
# (the error case doesn't include the msgs because otherwise they'd
# appear twice, once now and once on stderr)
        syntax_error_details = StringIO()
        r = pyflakes.reporter.Reporter(syntax_error_details, syntax_error_details)
        r.syntaxError(filename, msg, lineno, offset, text)
# return more info
        return {'message_for_onlywarn': syntax_error_details.getvalue(),
                'messages':["Error - see captured stderr, below."]}
###########################################################################
    except Exception:
        reporter.unexpectedError(filename, 'problem decoding source')
### nbsmoke addition ######################################################
# return more info
        return {'message_for_onlywarn': 'problem decoding source',                
                'messages':["Error - see captured stderr, below."]}
###########################################################################
    # Okay, it's syntactically valid.  Now check it.
    w = pyflakes.checker.Checker(tree, filename)

### nbsmoke addition ######################################################
# hack to support '# noqa' in ipynb
    NOQA = re.compile('# noqa', re.IGNORECASE)
    noqa_lines = [i+1 for i,l in enumerate(codeString.splitlines()) if NOQA.search(l)]
    # (I assume there was a reason for updating list in place?)
    w.messages[:] = [m for m in w.messages if m.lineno not in noqa_lines]
###########################################################################

### nbsmoke addition ######################################################
# hack to support ' # nbsmoke-blacklisted: <magic>'
    blacklist = re.compile('.* # nbsmoke-blacklisted: (.*)', re.IGNORECASE)
    blacklist_messages = []
    for i,l in enumerate(codeString.splitlines()):
        m = blacklist.match(l)
        if m:
            blacklist_messages.append(
                FakeMessage(lineno=i+1, col=m.start(1), message="nbsmoke blacklisted magic: %s", message_args=(m.group(1),)))
    w.messages[:] = w.messages + blacklist_messages
###########################################################################

    w.messages.sort(key=lambda m: m.lineno)

### nbsmoke addition ######################################################
# return more info
    return {'messages':[("line %s col %s: "%(msg.lineno,msg.col))+msg.message%msg.message_args for msg in w.messages]}
###########################################################################



# -*- coding: utf-8 -*-

import os

from . import nb_basic, WARNINGS_ARE_ERRORS, lint_args


def test_lint_good(testdir):
    testdir.makefile('.ipynb', testing123=nb_basic%{'the_source':"1/1"})
    result = testdir.runpytest(*lint_args)
    assert result.ret == 0

def test_lint_bad_syntax(testdir):
    testdir.makefile('.ipynb', testing123=nb_basic%{'the_source':"1/1 these undefined names are a syntax error"})
    result = testdir.runpytest(*lint_args)
    assert result.ret == 1
    assert result.parseoutcomes()['failed'] == 1
    result.stdout.re_match_lines_random(
        [".*invalid syntax.*",
         ".*1/1 these undefined names are a syntax error.*"])

def test_lint_bad(testdir):
    testdir.makefile('.ipynb', testing123=nb_basic%{'the_source':"undefined_name"})
    result = testdir.runpytest(*lint_args)
    assert result.ret == 1
    assert result.parseoutcomes()['failed'] == 1
    result.stdout.re_match_lines_random(
        [".*undefined name 'undefined_name'.*",
         ".*To see python source that was checked by pyflakes.*pass --nbsmoke-lint-debug$",
         ".*To see python source before magics were handled by nbsmoke.*pass --nbsmoke-lint-debug$"])


def test_lint_bad_silenced_with_noqa(testdir):
    testdir.makefile('.ipynb', testing123=nb_basic%{'the_source':"undefined # noqa: deliberate :)"})
    result = testdir.runpytest('--nbsmoke-lint','-v')
    assert result.ret == 0

def test_lint_bad_onlywarn(testdir):
    testdir.makefile('.ipynb', testing123=nb_basic%{'the_source':"1/1 these undefined names are definitely undefined"})
    _args = list(lint_args)
    _args.remove(WARNINGS_ARE_ERRORS)
    _args.append("--nbsmoke-lint-onlywarn")
    result = testdir.runpytest(*_args)
    assert result.ret == 0
    # seems like this changed in pytest or something at some point!
    warningkey = 'warnings' if 'warnings' in result.parseoutcomes() else 'warning'
    assert result.parseoutcomes()[warningkey] == 1
    result.stdout.re_match_lines_random(
        [".*invalid syntax.*",
         ".*1/1 these undefined names are definitely undefined.*"])

def test_lint_bad_debug(testdir):
    testdir.makefile('.ipynb', testing123=nb_basic%{'the_source':"undefined_name"})
    result = testdir.runpytest(*(lint_args + ['--nbsmoke-lint-debug']))

    assert result.ret == 1
    assert result.parseoutcomes()['failed'] == 1
    result.stdout.re_match_lines_random(
        [".*undefined name 'undefined_name'.*",
         ".*To see python source that was checked by pyflakes.*testing123.nbsmoke-debug-postmagicprocess.py$",
         ".*To see python source before magics were handled by nbsmoke.*testing123.nbsmoke-debug-premagicprocess.py$"])
    assert os.path.isfile("testing123.nbsmoke-debug-premagicprocess.py")
    assert os.path.isfile("testing123.nbsmoke-debug-postmagicprocess.py")

def test_flakes_to_ignore(testdir):
    testdir.makeini(r"""
        [pytest]
        nbsmoke_flakes_to_ignore = .*hvplot.* imported but unused.*
    """)
    testdir.makefile('.ipynb', testing123=nb_basic%{'the_source':"import hvplot.pandas; import hvplot.matplotlib"})
    result = testdir.runpytest(*lint_args)
    assert result.ret == 0



# -*- coding: utf-8 -*-

import os
import codecs
import json
from setuptools import setup, find_packages


def get_setup_version(reponame):
    """
    Helper to get the current version from either git describe or the
    .version file (if available).
    """
    basepath = os.path.split(__file__)[0]
    version_file_path = os.path.join(basepath, reponame, '.version')
    try:
        from param import version
    except Exception:
        version = None
    if version is not None:
        return version.Version.setup_version(basepath, reponame, archive_commit="$Format:%h$")
    else:
        print("WARNING: param>=1.6.0 unavailable. If you are installing a package, "
              "this warning can safely be ignored. If you are creating a package or "
              "otherwise operating in a git repository, you should install param>=1.6.0.")
        return json.load(open(version_file_path, 'r'))['version_string']


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding='utf-8').read()

extras_require = {
    'holoviews-magics': ['holoviews'], # only install if you are using holoviews magics (which are deprecated...)
    'verify': ['requests', 'beautifulsoup4'],
    # until pyproject.toml/equivalent is widely supported (setup_requires
    # doesn't work well with pip)
    'build': [
        'param>=1.7.0',
        'pyct>=0.4.4',
        'setuptools>=30.3.0',
    ],
    'lint': ['flake8'],
    'tests': [
        'coverage',
        'twine',   # required for pip packaging
        'rfc3986', # required by twine
        'keyring', # required by twine
    ],
}

extras_require['tests'] += extras_require['lint']
extras_require['tests'] += extras_require['holoviews-magics']
extras_require['tests'] += extras_require['verify']

setup_args = dict(
    name='nbsmoke',
    description='Basic notebook checks. Do they run? Do they contain lint?',
    version = get_setup_version('nbsmoke'),
    url='https://github.com/pyviz-dev/nbsmoke',
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    author='pyviz contributors',
    author_email = "dev@pyviz.org",
    license='BSD-3',
    packages=find_packages(),
    include_package_data = True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: BSD License',
    ],
    
    python_requires = ">=3.4",
    
    install_requires=[
        'param>=1.7.0',
        'pytest>=3.1.1',
        ########## lint stuff
        'pyflakes',
        ########## notebook stuff (reading, executing)
        # * Required imports: nbconvert, nbformat.
        # * Need to be able to execute ipython notebooks.
        # * Optional: process ipython magics (required import: IPython)
        'jupyter_client',
        'nbformat',
        'nbconvert',
        'ipykernel'],
    extras_require = extras_require,
    entry_points={
        'pytest11': [
            'nbsmoke = nbsmoke',
        ],
    },
)

setup_args['extras_require']['all'] = sorted(set(sum(extras_require.values(), [])))

if __name__=="__main__":
    setup(**setup_args)



import pytest
holoviews = pytest.importorskip("holoviews", reason="HoloViews is required to test hv magics.")

from . import lint_args

# otherwise stuff from the bowels of pandas and numpy comes up
lint_args = lint_args + ['-W ignore::RuntimeWarning']

nb_hv_good = u'''
{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import holoviews as hv\\n",
    "import os\\n",
    "from holoviews.operation.datashader import datashade, regrid"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%opts RGB [invert_yaxis=True width=400 height=400]\\n",
    "%opts HeatMap (cmap='Winter') [colorbar=True, toolbar='random']"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%opts HeatMap (cmap='Summer') [colorbar=True, toolbar='above']\\n",
    "%opts HeatMap (cmap='Summer') [colorbar=True, toolbar='blinking']"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%opts RGB [invert_yaxis=True width=400 height=400]\\n",
    "adim = hv.Dimension('A')\\n",
    "bdim = hv.Dimension('B')\\n",
    "paths = [os.path.join('data', f) for f in ['a','b']]\\n",
    "hv.Layout([datashade(hv.Curve(p, kdims=[adim], vdims=[bdim])) for p in paths]).cols(2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%opts Image [width=500 height=500 colorbar=True colorbar_position='left' logz=True]\\n",
    "%%opts Image [xticks=[1,2] yticks=[2,3]] (cmap='viridis')\\n",
    "%%opts HLine (color='red' line_width=2) VLine (color='red', line_width=2)\\n",
    "im = hv.Image({'a':1, 'b':2, 'c':3}, kdims=[adim, bdim], \\n",
    "              vdims=[hv.Dimension('c')])\\n",
    "regrid(im) * regrid(im)"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python",
   "pygments_lexer": "ipython3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
'''

nb_hv_bad = u'''
{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import holoviews as hv\\n",
    "import os\\n",
    "from holoviews.operation.datashader import datashade"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%opts RGB [invert_yaxis=True width=400 height=400]\\n",
    "adim = hv.Dimension('A')\\n",
    "bdim = hv.Dimension('B')\\n",
    "paths = [os.path.join('data', f) for f in ['a','b']]\\n",
    "hv.Layout([datashade(hv.Curve(no, kdims=[adim], vdims=[bdim])) for p in paths]).cols(2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%opts Image [width=500 height=500 colorbar=True colorbar_position='left' logz=True]\\n",
    "%%opts Image [xticks=xticks yticks=yticks] (cmap='viridis')\\n",
    "%%opts HLine (color='red' line_width=2) VLine (color='red', line_width=2)\\n",
    "im = hv.Image({'a':a, 'b':b}, kdims=['a'], \\n",
    "              vdims=[hv.Dimension('b')])\\n",
    "regrid(im) * lines(no)"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python",
   "pygments_lexer": "ipython3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
'''

# note: % escaped as % subst
nb_cell_and_line = u'''
{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import holoviews as hv\\n",
    "import os\\n",
    "from holoviews.operation.datashader import datashade"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%%%opts HLine (color='red' line_width=2) VLine (color='red', line_width=2)\\n",
    "%%%%opts RGB [invert_yaxis=True width=400 height=400]\\n",
    "%%time %(dim_name)s = hv.Dimension('A')\\n",
    "bdim = hv.Dimension('B')\\n",
    "paths = [os.path.join('data', f) for f in ['a','b']]\\n",
    "hv.Layout([datashade(hv.Curve(p, kdims=[adim], vdims=[bdim])) for p in paths]).cols(2)"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python",
   "pygments_lexer": "ipython3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
'''

nb_bad_opts_syntax = u'''
{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import holoviews as hv"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%opts RGB bad syntax is bad [invert_yaxis=True width=400 height=400]\\n",
    "hv.Dimension('B')"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python",
   "pygments_lexer": "ipython3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
'''

nb_bad_opts_syntax2 = u'''
{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import holoviews as hv"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%opts RGB [invert_yaxis=True width=400 height=400]should not be here\\n",
    "hv.Dimension('B')"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python",
   "pygments_lexer": "ipython3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
'''


def test_lint_magics_hv_good(testdir):
    # TODO: right now just that it hasn't caused errors, but should
    # check there's content + noqa in there
    testdir.makefile('.ipynb', testing123=nb_hv_good)
    result = testdir.runpytest(*lint_args)
    assert result.ret == 0

def test_lint_magics_hv_bad(testdir):
    testdir.makefile('.ipynb', testing123=nb_hv_bad)
    result = testdir.runpytest(*lint_args)
    result.stdout.re_match_lines_random(
        [".*undefined name 'no'$",
         ".*undefined name 'a'$",
         ".*undefined name 'b'$",
         ".*undefined name 'regrid'$",
         ".*undefined name 'lines'$",
         ".*undefined name 'no'$"])
    assert result.ret == 1

def test_lint_magics_hv_cell_and_line_good(testdir):
    testdir.makefile('.ipynb', testing123=nb_cell_and_line%{'dim_name':'adim'})
    result = testdir.runpytest(*lint_args)
    assert result.ret == 0

def test_lint_magics_hv_cell_and_line_bad(testdir):
    testdir.makefile('.ipynb', testing123=nb_cell_and_line%{'dim_name':'bad_name'})
    result = testdir.runpytest(*lint_args)
    result.stdout.re_match_lines_random(
        [".*undefined name 'adim'$"])
    assert result.ret == 1

def test_lint_magics_hv_bad_opts_syntax(testdir):
    testdir.makefile('.ipynb', testing123=nb_bad_opts_syntax)
    result = testdir.runpytest(*lint_args)
    assert result.ret == 1
    result.stdout.re_match_lines_random(
        ["^E.*SyntaxError: Invalid specification syntax.$"])

def test_lint_magics_hv_bad_opts_syntax2(testdir):
    testdir.makefile('.ipynb', testing123=nb_bad_opts_syntax2)
    result = testdir.runpytest(*lint_args)
    assert result.ret == 1
    result.stdout.re_match_lines_random(
        ["^E.*SyntaxError: Failed to parse remainder of string: u?'should not be here'"])
